#include <stdio.h>

#include <GL/glew.h>
#include <GLFW/glfw3.h>

#include "vertexShader.c"
#include "fragmentShader.c"

#include "matrixMath.h"
#include "transformation.h"
#include "wavefrontobj.h"

#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <stdbool.h>

#define RESTART 345678

GLuint program;
GLuint vao;

int numFaces = 0; 

bool exitRequested = false;

GLFWwindow* window;

GLfloat aspectRatio = 1.0f;

double timeBetweenUpdates = 0.2f;

double timeSinceUpdate = 0.0f;
int framesSinceUpdate = 0;

GLfloat step = 0.0f;
const GLfloat pi = 3.14159f;

vec3 cameraPosition = {0.0f, 3.0f, 5.5f};

// input handler for camera movement
void handleInputs(double deltaTime) {
	if (glfwGetKey(window, GLFW_KEY_S) == GLFW_PRESS) {
		cameraPosition.z += deltaTime * 10;
	}
	if (glfwGetKey(window, GLFW_KEY_W) == GLFW_PRESS) {
		cameraPosition.z -= deltaTime * 10;
	}
	if (glfwGetKey(window, GLFW_KEY_SPACE) == GLFW_PRESS) {
		cameraPosition.y += deltaTime * 10;
	}
	if (glfwGetKey(window, GLFW_KEY_LEFT_SHIFT) == GLFW_PRESS) {
		cameraPosition.y -= deltaTime * 10;
	}
}

// input handler to quit with ESC
void keyboardHandler(GLFWwindow* window, int key, int scancode, int action, int mods) {
	if (action == GLFW_PRESS) {
		if (key == GLFW_KEY_ESCAPE) {
			exitRequested = true;
		}
	}
}


void init(void) {
	// create and compile vertex shader
	const GLchar *vertexTextConst = vertexShader_glsl;

	GLuint vertexShader = glCreateShader(GL_VERTEX_SHADER);
	glShaderSource(vertexShader, 1, &vertexTextConst, &vertexShader_glsl_len);
	glCompileShader(vertexShader);


	GLint status;
	glGetShaderiv(vertexShader, GL_COMPILE_STATUS, &status);
	
	if (!status) {
		printf("Error compiling vertex shader: ");
		GLchar infoLog[1024];
		glGetShaderInfoLog(vertexShader, 1024, NULL, infoLog);
		printf("%s",infoLog);
	}

	vertexTextConst = NULL;



	// create and compile fragment shader
	
	const GLchar *fragmentTextConst = fragmentShader_glsl;

	GLuint fragmentShader = glCreateShader(GL_FRAGMENT_SHADER);
	glShaderSource(fragmentShader, 1, &fragmentTextConst, &fragmentShader_glsl_len);
	glCompileShader(fragmentShader);

	glGetShaderiv(fragmentShader, GL_COMPILE_STATUS, &status);
	
	if (!status) {
		printf("Error compiling fragment shader: ");
		GLchar infoLog[1024];
		glGetShaderInfoLog(fragmentShader, 1024, NULL, infoLog);
		printf("%s",infoLog);
	}

	// create and link shader program
	program = glCreateProgram();
	glAttachShader(program, vertexShader);
	glAttachShader(program, fragmentShader);
	glLinkProgram(program);

	glGetProgramiv(program, GL_LINK_STATUS, &status);

	if (!status) {
		printf("Error linking program: ");
		GLchar infoLog[1024];
		glGetProgramInfoLog(program, 1024, NULL, infoLog);
		printf("%s",infoLog);
	}
	glValidateProgram(program);


	glGetProgramiv(program, GL_VALIDATE_STATUS, &status);

	if (!status) {
		printf("Error validating program: ");
		GLchar infoLog[1024];
		glGetProgramInfoLog(program, 1024, NULL, infoLog);
		printf("%s",infoLog);
	}


	// --------------- READ teapot.obj
	ParsedObjFile teapot = readObjFile("../obj/monkey.obj");
	numFaces = teapot.length;

	// write teapot faces to buffer
	GLuint triangleVertexBufferObject;
	glGenBuffers(1, &triangleVertexBufferObject);
	glBindBuffer(GL_ARRAY_BUFFER, triangleVertexBufferObject);
	glBufferData(GL_ARRAY_BUFFER, teapot.length * sizeof(face), teapot.faces, GL_STATIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, 0);


	// create vertex array object
	glGenVertexArrays(1, &vao);
	glBindVertexArray(vao);
	glBindBuffer(GL_ARRAY_BUFFER, triangleVertexBufferObject);

	// vertex positions
	glVertexAttribPointer(
		0,
		3,
		GL_FLOAT,
		GL_FALSE,
		sizeof(vertex),
		0
	);
	glEnableVertexAttribArray(0);

	// vertex normals
	glVertexAttribPointer(
		1,
		3,
		GL_FLOAT,
		GL_FALSE,
		sizeof(vertex),
		(void*) sizeof(vec3)
	);
	glEnableVertexAttribArray(1);

	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindVertexArray(0);


	// ENABLE BACKFACE CULLING
	glFrontFace(GL_CCW);
	glEnable(GL_CULL_FACE);

	// ENABLE DEPTH BUFFER
	glEnable(GL_DEPTH_TEST);

	glClearColor(0.1f, 0.1f, 0.1f, 1.0f);
}

void updateStats() {
	printf("\rFPS: %.1f", framesSinceUpdate / timeSinceUpdate);
	printf(" - Camera Position: [%f, %f, %f]", cameraPosition.x, cameraPosition.y, cameraPosition.z);
	fflush(stdout);
}

void draw(void) {

	// FPS Counter
	framesSinceUpdate++;
	double deltaTime = glfwGetTime();
	timeSinceUpdate += deltaTime;
	glfwSetTime(0.0f);

	if (timeSinceUpdate >= timeBetweenUpdates) {
		updateStats();
		timeSinceUpdate = 0.0f;
		framesSinceUpdate = 0;
	}

	// camera movement
	handleInputs(deltaTime);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glUseProgram(program);
	glBindVertexArray(vao);

	// step for rotations
	// counts up to 1.0 and then resets back to 0.0 forever
	step += deltaTime / 5;
	if (step > 1.0f) step -= 1.0f;

	// step multiplied by pi * 2 for use in rotation and trig functions
	GLfloat stepi = step * pi * 2;


	// ------------- MODEL TRANSFORMATION ---------------------
	// SCALE -> ROTATE -> TRANSLATE

	mat4 modelTransformation;
	identity(&modelTransformation);

	rotateY(&modelTransformation, &modelTransformation, stepi);


	// ------------- VIEWING TRANSFORMATION -------------------
	vec3 origin = {0.0f, 0.0f, 0.0f};
	vec3 up = {0.0f, 1.0f, 0.0f};

	mat4 viewingTransformation;
	lookAt(&viewingTransformation, &cameraPosition, &origin, &up);



	// -------------- PROJECTION TRANSFORMATION ----------------
	mat4 projectionTransformation;
	GLfloat near = 0.1f;
	GLfloat far = 20.0f;
	perspectiveProjection(&projectionTransformation, near, far);



	// -------------- NORMALISATION TRANSFORMATION -------------
	mat4 normalisationTransformation;
	GLfloat fovy = pi / 2;
	normalisedDeviceCoordinatesFov(&normalisationTransformation, fovy, aspectRatio, near, far);


	mat4 modelView;
	identity(&modelView);
	multiply(&modelView, &modelTransformation, &modelView);
	multiply(&modelView, &viewingTransformation, &modelView);

	mat4 projection;
	identity(&projection);
	multiply(&projection, &projectionTransformation, &projection);
	multiply(&projection, &normalisationTransformation, &projection);


	// send transformation matrix to shader
	glUniformMatrix4fv(glGetUniformLocation(program, "modelView"), 1, GL_FALSE, (GLfloat*)&modelView);
	glUniformMatrix4fv(glGetUniformLocation(program, "projection"), 1, GL_FALSE, (GLfloat*)&projection);


	vec4 lightPosition = {cos(stepi * 2) * 3.0f, 3.0f, sin(stepi * 2) * 3.0f, 1.0f};
	multiplyAny((GLfloat *)&lightPosition, (GLfloat *)&modelView, (GLfloat *)&lightPosition, 4, 4, 1);

	glUniform3f(glGetUniformLocation(program, "lightPosition"), lightPosition.x, lightPosition.y, lightPosition.z);

	// SET MATERIAL DATA
	glUniform4f(glGetUniformLocation(program, "ambientColor"), 0.25f, 0.22f, 0.06f, 1.0f);
	glUniform4f(glGetUniformLocation(program, "diffusionColor"), 0.35f, 0.31f, 0.09f, 1.0f);
	glUniform4f(glGetUniformLocation(program, "specularColor"), 0.80f, 0.72f, 0.21f, 1.0f);
	glUniform1f(glGetUniformLocation(program, "shininess"), 83.2f * 4.0f);

	// SET LIGHT DATA
	glUniform4f(glGetUniformLocation(program, "lightColor"), 1.0f, 1.0f, 1.0f, 1.0f);
	glUniform4f(glGetUniformLocation(program, "ambientLight"), 0.2f, 0.2f, 0.2f, 1.0f);

	// draw!!1
	glDrawArrays(GL_TRIANGLES, 0, numFaces * 3);
}

// change viewport size and adjust aspect ratio when changing window size
void framebuffer_size_callback(GLFWwindow *window, int width, int height) {
	glViewport(0, 0, width, height);
	aspectRatio = (float)width / height;
}

int main(void) {

	// initialise window
	glfwInit();
	glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
	glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
	glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);

	window = glfwCreateWindow(700, 700, "Computergrafik 1", NULL, NULL);

	if (!window) {
		printf("Failed to create window\n");
		glfwTerminate();
		return -1;
	}


	glfwSetFramebufferSizeCallback(window, framebuffer_size_callback);
	glfwMakeContextCurrent(window);

	// disable framerate cap
	glfwSwapInterval(0);

	// register keyboard event handler
	glfwSetKeyCallback(window, keyboardHandler);

	// initialise glew
	glewInit();

	printf("OpenGL version supported by this platform (%s):\n", glGetString(GL_VERSION));

	init();

	// exit when window should close or exit is requested (ESC)
	while (!glfwWindowShouldClose(window) && !exitRequested) {
		draw();

		glfwSwapBuffers(window);
		glfwPollEvents();
	}

	glfwTerminate();

	return 0;
}